'''
Testing Ragaraja Interpreter
Date created: 20th August 2012
Licence: Python Software Foundation License version 2

The tests are executed in sequence from test 1 to test N where the test N 
specific code is added on the the concatenated codes/instructions from the
previous test in the form of

code(N) = code(N-1) + code(N)

Although it is possible to use unittest, it is probably mind-boggling to
try to understand long stretches of codes. It is probably easier to gradually
build up on the code and knowledge/state of the previous instructions. This
means that it is not possible to pass test N but fail test N-1.

Sample test case/scenario capsule -

1: {'restart': False                
    'in_source': '008008008008',    
    'forcedinarray': [0]*10
    'array': [4,0,0,0,0,0,0,0,0,0], 
    'apointer': 0,
    'forcedindata': []
    'inputdata': [],
    'output': [],
    'out_source': '008008008008',
    'spointer': 12,
    <register>
   }
   
containing 10 options:
1. 'restart' - determines if the tester should clear concatenated instructions 
    so far. Default = False, which means keep the clear concatenated 
    instructions so far and add on the current instructions to execute.
    Not used unless to clean out concatenated instructions so far.
2. 'in_source' - the list of specific instructions to execute in this test,
    which will be appended to the list of concatenated instructions so far.
3. 'forcedinarray' - array/tape to be fed into register machine before code
    execution (code = concatenated instructions so far + in_source). If not
    defined, it will be set to [0,0,0,0,0,0,0,0,0,0].
4. 'array' - array/tape at the end of execution.
5. 'apointer' - array pointer at the end of execution.
6. 'forcedindata' - input data list to be fed into register machine before 
    code execution (code = concatenated instructions so far + in_source). 
    If not defined, it will be set to [].
7. 'inputdata' - input data list (defined by 'forcedindata') at the end of
    execution.
8. 'output' - output list at the end of execution.
9. 'out_source': concatenated instructions so far + in_source.
10. 'spointer': source pointer at the end of execution on 'out_source'.
11. 'register': a toggle switch (if present) to print internal register.
'''

import sys
import os

sys.path.append(os.path.join(os.path.dirname(os.getcwd()), 'copads'))
import ragaraja as N

import register_machine as r

testdata = {
            # Tests 1-7 are the same tests as for NucleotideBF (t_n_bf.py)
            # Testing 000, 004, 008, 011
            1: {'in_source': '008008008008',
                'array': [4,0,0,0,0,0,0,0,0,0],
                'apointer': 0,
                'inputdata': [],
                'output': [],
                'out_source': '008008008008',
                'spointer': 12},
            2: {'in_source': '000000',
                'array': [4,0,0,0,0,0,0,0,0,0],
                'apointer': 2,
                'inputdata': [],
                'output': [],
                'out_source': '008008008008000000',
                'spointer': 18},
            3: {'in_source': '011011011',
                'array': [4,0,-3,0,0,0,0,0,0,0],
                'apointer': 2,
                'inputdata': [],
                'output': [],
                'out_source': '008008008008000000011011011',
                'spointer': 27},
            4: {'in_source': '004004004004',
                'array': [4,0,-3,0,0,0,0,0,0,0],
                'apointer': 8,
                'inputdata': [],
                'output': [],
                'out_source': '008008008008000000011011011004004004004',
                'spointer': 39},
            5: {'in_source': '011011011011011',
                'array': [4,0,-3,0,0,0,0,0,-5,0],
                'apointer': 8,
                'inputdata': [],
                'output': [],
                'out_source': '008008008008000000011011011004004004004011011011\
011011',
                'spointer': 54},
            6: {'in_source': '000000000',
                'array': [4,0,-3,0,0,0,0,0,-5,0],
                'apointer': 1,
                'inputdata': [],
                'output': [],
                'out_source': '008008008008000000011011011004004004004011011011\
011011000000000',
                'spointer': 63},
            7: {'in_source': '008008008',
                'array': [4,3,-3,0,0,0,0,0,-5,0],
                'apointer': 1,
                'inputdata': [],
                'output': [],
                'out_source': '008008008008000000011011011004004004004011011011\
011011000000000008008008',
                'spointer': 72},
            # Testing 001, 009
            8: {'in_source': '001009',
                'array': [4,3,-3,0,0,0,5,0,-5,0],
                'apointer': 6,
                'inputdata': [],
                'output': [],
                'out_source': '008008008008000000011011011004004004004011011011\
011011000000000008008008001009',
                'spointer': 78},
            # Testing 002, 005
            9: {'in_source': '002005010',
                'array': [4,13,-3,0,0,0,5,0,-5,0],
                'apointer': 1,
                'inputdata': [],
                'output': [],
                'out_source': '008008008008000000011011011004004004004011011011\
011011000000000008008008001009002005010',
                'spointer': 87},
            # Testing 003, 043
            10: {'in_source': '043003',
                 'array': [4,13,-3,0,0,0,5,0,-5,0],
                 'apointer': 6,
                 'inputdata': [],
                 'output': [],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003',
                 'spointer': 93},
            # Testing 006, 044
            11: {'in_source': '044006',
                 'array': [4,13,-3,0,0,0,5,0,-5,0],
                 'apointer': 9,
                 'inputdata': [],
                 'output': [],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006',
                 'spointer': 99},
            # Testing 012, 013, 032, 033
            12: {'in_source': '012032013033',
                 'array': [4,13,-3,0,0,0,5,0,-5,-10],
                 'apointer': 9,
                 'inputdata': [],
                 'output': [],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033',
                 'spointer': 111},
            # Testing 061, 062
            13: {'in_source': '004004004061004062',
                 'array': [4,13,-3,0,0,0,5,0,-5,-10],
                 'apointer': 6,
                 'inputdata': [],
                 'output': [],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
062',
                 'spointer': 129},
            # Testing 016, 018
            14: {'in_source': '016018',
                 'array': [4,13,-3,0,0,0,5,0,-5,-10],
                 'apointer': 6,
                 'inputdata': [],
                 'output': [],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
062016018',
                 'spointer': 135},
            # Testing 017, 019
            15: {'in_source': '017019',
                 'array': [4,13,-3,0,0,0,5,0,-5,-10],
                 'apointer': 6,
                 'inputdata': [],
                 'output': [],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
062016018017019',
                 'spointer': 141},
            # Testing 046
            16: {'in_source': '046046',
                 'array': [4,13,-3,0,0,0,5,0,-5,-10],
                 'apointer': 6,
                 'inputdata': [],
                 'output': [],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
062016018017019046046',
                 'spointer': 147},
            # Testing 081, 133
            17: {'in_source': '081133',
                 'array': [4,13,-3,0,0,0,0,-10,-5,5],
                 'apointer': 6,
                 'inputdata': [],
                 'output': [],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
062016018017019046046081133',
                 'spointer': 153},
            # Testing 084, 085, 086
            18: {'in_source': '000084000085000086000097000098',
                 'array': [3.14159265358979323846,2.718281828459045,
                           -3,0,0,0,0,0,-1,1],
                 'apointer': 1,
                 'inputdata': [],
                 'output': [],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
062016018017019046046081133000084000085000086000097000098',
                 'spointer': 183},
            # Testing 120, 121, 122
            19: {'in_source': '004120000121000122',
                 'array': [1,1,1,0,0,0,0,0,-1,1],
                 'apointer': 2,
                 'inputdata': [],
                 'output': [],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
062016018017019046046081133000084000085000086000097000098004120000121000122',
                 'spointer': 201},
            # Testing 065, 068, 071, 115
            20: {'in_source': '004004008004008071065068000008115',
                 'array': [6,1,1,0,0,0,0,0,-1,-4],
                 'apointer': 0,
                 'inputdata': [],
                 'output': [],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
04008004008071065068000008115',
                 'spointer': 234},
            # Testing 021, 022, 038, 042, 047
            21: {'in_source': '021022047000038042',
                 'array': [6,0,1,0,0,0,0,0,-1,-4],
                 'apointer': 1,
                 'inputdata': [],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
04008004008071065068000008115021022047000038042',
                 'spointer': 252},
            # Testing 080, 116
            22: {'in_source': '009004116080',
                 'array': [7,5,1,0,0,0,0,0,-1,-4],
                 'apointer': 0,
                 'inputdata': [],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
04008004008071065068000008115021022047000038042009004116080',
                 'spointer': 264},
            # Testing 077, 087, 089, 088, 090, 094, 095, 096, 105, 108, 109, 117
            23: {'in_source': '077087009089088090094108096109117105095080',
                 'array': [8,5,1,0,0,0,0,0,-1,-4],
                 'apointer': 0,
                 'inputdata': [],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080',
                 'spointer': 306},
            # Testing 091, 092, 093, 099, 100, 101, 102, 103, 104, 112, 113
            24: {'in_source': '000000091093099101092100102103104112113105080',
                 'array': [8,5,50,0,0,0,0,0,-1,-4],
                 'apointer': 2,
                 'inputdata': [],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080',
                 'spointer': 351},
            # Pushing [1,2,3,4,5,6,7,8,9] as input data into the test system
            25: {'in_source': '',
                 'array': [8,5,50,0,0,0,0,0,-1,-4],
                 'apointer': 2,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080',
                 'spointer': 351},
            # Testing 066, 067, 069, 070, 072, 073, 075, 076
            26: {'in_source': '066000067000069000070005000072073076075',
                 'array': [8,5,51,9,1,9,0,0,-1,-4],
                 'apointer': 1,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
069000070005000072073076075',
                 'spointer': 390},
            # Testing 140, 141, 142
            27: {'in_source': '140008141013142008',
                 'array': [8,5,41,9,2,9,1,0,-1,-4],
                 'apointer': 6,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
069000070005000072073076075140008141013142008',
                 'spointer': 408},
            # Testing 189, 201, 202, 203, 204, 205, 206, 207, 208, 209, 210, 
            # 301, 302, 303, 304, 305, 306, 307, 308, 309, 310
            28: {'in_source': '004005201000202000203000204000205000206000207000\
2080002090002100001893100003090003080003070003060003050003040003030003020003010\
00',
                 'array': [-4,-1,0,1,9,2,9,41,5,8],
                 'apointer': 0,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
0690000700050000720730760751400081410131420080040052010002020002030002040002050\
0020600020700020800020900021000018931000030900030800030700030600030500030400030\
3000302000301000',
                 'spointer': 537},
            # Testing 211, 212, 213, 214, 215, 216, 217, 218, 219, 220, 
            # 311, 312, 313, 314, 315, 316, 317, 318, 319, 320
            29: {'in_source': '211000212000213000214000215000216000217000218000\
219000220000320000319000318000317000316000315000314000313000312000311000',
                 'array': [8,5,41,9,2,9,1,0,-1,-4],
                 'apointer': 0,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
0690000700050000720730760751400081410131420080040052010002020002030002040002050\
0020600020700020800020900021000018931000030900030800030700030600030500030400030\
3000302000301000211000212000213000214000215000216000217000218000219000220000320\
000319000318000317000316000315000314000313000312000311000',
                 'spointer': 657},
            # Testing 221, 222, 223, 224, 225, 226, 227, 228, 229, 230, 
            # 321, 322, 323, 324, 325, 326, 327, 328, 329, 330
            30: {'in_source': '221000222000223000224000225000226000227000228000\
229000230000330000329000328000327000326000325000324000323000322000321000',
                 'array': [-4,-1,0,1,9,2,9,41,5,8],
                 'apointer': 0,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
0690000700050000720730760751400081410131420080040052010002020002030002040002050\
0020600020700020800020900021000018931000030900030800030700030600030500030400030\
3000302000301000211000212000213000214000215000216000217000218000219000220000320\
0003190003180003170003160003150003140003130003120003110002210002220002230002240\
0022500022600022700022800022900023000033000032900032800032700032600032500032400\
0323000322000321000',
                 'spointer': 777},
            # Testing 231, 232, 233, 234, 235, 236, 237, 238, 239, 240, 
            # 331, 332, 333, 334, 335, 336, 337, 338, 339, 340
            31: {'in_source': '231000232000233000234000235000236000237000238000\
239000240000340000339000338000337000336000335000334000333000332000331000',
                 'array': [8,5,41,9,2,9,1,0,-1,-4],
                 'apointer': 0,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
0690000700050000720730760751400081410131420080040052010002020002030002040002050\
0020600020700020800020900021000018931000030900030800030700030600030500030400030\
3000302000301000211000212000213000214000215000216000217000218000219000220000320\
0003190003180003170003160003150003140003130003120003110002210002220002230002240\
0022500022600022700022800022900023000033000032900032800032700032600032500032400\
0323000322000321000231000232000233000234000235000236000237000238000239000240000\
340000339000338000337000336000335000334000333000332000331000',
                 'spointer': 897},
            # Testing 241, 242, 243, 244, 245, 246, 247, 248, 249, 250, 
            # 341, 342, 343, 344, 345, 346, 347, 348, 349, 350
            32: {'in_source': '241000242000243000244000245000246000247000248000\
249000250000350000349000348000347000346000345000344000343000342000341000',
                 'array': [-4,-1,0,1,9,2,9,41,5,8],
                 'apointer': 0,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
0690000700050000720730760751400081410131420080040052010002020002030002040002050\
0020600020700020800020900021000018931000030900030800030700030600030500030400030\
3000302000301000211000212000213000214000215000216000217000218000219000220000320\
0003190003180003170003160003150003140003130003120003110002210002220002230002240\
0022500022600022700022800022900023000033000032900032800032700032600032500032400\
0323000322000321000231000232000233000234000235000236000237000238000239000240000\
3400003390003380003370003360003350003340003330003320003310002410002420002430002\
4400024500024600024700024800024900025000035000034900034800034700034600034500034\
4000343000342000341000',
                 'spointer': 1017},
            # Testing 251, 252, 253, 254, 255, 256, 257, 258, 259, 260, 
            # 351, 352, 353, 354, 355, 356, 357, 358, 359, 360
            33: {'in_source': '251000252000253000254000255000256000257000258000\
259000260000360000359000358000357000356000355000354000353000352000351000',
                 'array': [8,5,41,9,2,9,1,0,-1,-4],
                 'apointer': 0,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
0690000700050000720730760751400081410131420080040052010002020002030002040002050\
0020600020700020800020900021000018931000030900030800030700030600030500030400030\
3000302000301000211000212000213000214000215000216000217000218000219000220000320\
0003190003180003170003160003150003140003130003120003110002210002220002230002240\
0022500022600022700022800022900023000033000032900032800032700032600032500032400\
0323000322000321000231000232000233000234000235000236000237000238000239000240000\
3400003390003380003370003360003350003340003330003320003310002410002420002430002\
4400024500024600024700024800024900025000035000034900034800034700034600034500034\
4000343000342000341000251000252000253000254000255000256000257000258000259000260\
000360000359000358000357000356000355000354000353000352000351000',
                 'spointer': 1137},
            # Testing 261, 262, 263, 264, 265, 266, 267, 268, 269, 270, 
            # 361, 362, 363, 364, 365, 366, 367, 368, 369, 370
            34: {'in_source': '261000262000263000264000265000266000267000268000\
269000270000370000369000368000367000366000365000364000363000362000361000',
                 'array': [-4,-1,0,1,9,2,9,41,5,8],
                 'apointer': 0,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
0690000700050000720730760751400081410131420080040052010002020002030002040002050\
0020600020700020800020900021000018931000030900030800030700030600030500030400030\
3000302000301000211000212000213000214000215000216000217000218000219000220000320\
0003190003180003170003160003150003140003130003120003110002210002220002230002240\
0022500022600022700022800022900023000033000032900032800032700032600032500032400\
0323000322000321000231000232000233000234000235000236000237000238000239000240000\
3400003390003380003370003360003350003340003330003320003310002410002420002430002\
4400024500024600024700024800024900025000035000034900034800034700034600034500034\
4000343000342000341000251000252000253000254000255000256000257000258000259000260\
0003600003590003580003570003560003550003540003530003520003510002610002620002630\
0026400026500026600026700026800026900027000037000036900036800036700036600036500\
0364000363000362000361000',
                 'spointer': 1257},
            # Testing 271, 272, 273, 274, 275, 276, 277, 278, 279, 280, 
            # 371, 372, 373, 374, 375, 376, 377, 378, 379, 380
            35: {'in_source': '271000272000273000274000275000276000277000278000\
279000280000380000379000378000377000376000375000374000373000372000371000',
                 'array': [8,5,41,9,2,9,1,0,-1,-4],
                 'apointer': 0,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
0690000700050000720730760751400081410131420080040052010002020002030002040002050\
0020600020700020800020900021000018931000030900030800030700030600030500030400030\
3000302000301000211000212000213000214000215000216000217000218000219000220000320\
0003190003180003170003160003150003140003130003120003110002210002220002230002240\
0022500022600022700022800022900023000033000032900032800032700032600032500032400\
0323000322000321000231000232000233000234000235000236000237000238000239000240000\
3400003390003380003370003360003350003340003330003320003310002410002420002430002\
4400024500024600024700024800024900025000035000034900034800034700034600034500034\
4000343000342000341000251000252000253000254000255000256000257000258000259000260\
0003600003590003580003570003560003550003540003530003520003510002610002620002630\
0026400026500026600026700026800026900027000037000036900036800036700036600036500\
0364000363000362000361000271000272000273000274000275000276000277000278000279000\
280000380000379000378000377000376000375000374000373000372000371000',
                 'spointer': 1377},
            # Testing 281, 282, 283, 284, 285, 286, 287, 288, 289, 290, 
            # 381, 382, 383, 384, 385, 386, 387, 388, 389, 390
            36: {'in_source': '281000282000283000284000285000286000287000288000\
289000290000390000389000388000387000386000385000384000383000382000381000',
                 'array': [-4,-1,0,1,9,2,9,41,5,8],
                 'apointer': 0,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
0690000700050000720730760751400081410131420080040052010002020002030002040002050\
0020600020700020800020900021000018931000030900030800030700030600030500030400030\
3000302000301000211000212000213000214000215000216000217000218000219000220000320\
0003190003180003170003160003150003140003130003120003110002210002220002230002240\
0022500022600022700022800022900023000033000032900032800032700032600032500032400\
0323000322000321000231000232000233000234000235000236000237000238000239000240000\
3400003390003380003370003360003350003340003330003320003310002410002420002430002\
4400024500024600024700024800024900025000035000034900034800034700034600034500034\
4000343000342000341000251000252000253000254000255000256000257000258000259000260\
0003600003590003580003570003560003550003540003530003520003510002610002620002630\
0026400026500026600026700026800026900027000037000036900036800036700036600036500\
0364000363000362000361000271000272000273000274000275000276000277000278000279000\
2800003800003790003780003770003760003750003740003730003720003710002810002820002\
8300028400028500028600028700028800028900029000039000038900038800038700038600038\
5000384000383000382000381000',
                 'spointer': 1497},
            # Testing 291, 292, 293, 294, 295, 296, 297, 298, 299,
            # 391, 392, 393, 394, 395, 396, 397, 398, 399
            37: {'in_source': '291000292000293000294000295000296000297000298000\
299000000399000398000397000396000395000394000393000392000391',
                 'array': [5,41,9,2,9,1,0,-1,-4,8],
                 'apointer': 8,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
0690000700050000720730760751400081410131420080040052010002020002030002040002050\
0020600020700020800020900021000018931000030900030800030700030600030500030400030\
3000302000301000211000212000213000214000215000216000217000218000219000220000320\
0003190003180003170003160003150003140003130003120003110002210002220002230002240\
0022500022600022700022800022900023000033000032900032800032700032600032500032400\
0323000322000321000231000232000233000234000235000236000237000238000239000240000\
3400003390003380003370003360003350003340003330003320003310002410002420002430002\
4400024500024600024700024800024900025000035000034900034800034700034600034500034\
4000343000342000341000251000252000253000254000255000256000257000258000259000260\
0003600003590003580003570003560003550003540003530003520003510002610002620002630\
0026400026500026600026700026800026900027000037000036900036800036700036600036500\
0364000363000362000361000271000272000273000274000275000276000277000278000279000\
2800003800003790003780003770003760003750003740003730003720003710002810002820002\
8300028400028500028600028700028800028900029000039000038900038800038700038600038\
5000384000383000382000381000291000292000293000294000295000296000297000298000299\
000000399000398000397000396000395000394000393000392000391',
                 'spointer': 1605},
            # Testing 007, 020, 023, 024, 025, 034, 035, 036, 045, 063, 064
            38: {'in_source': '000007020023000000024001002003004005000025001002\
003004005006007008009010000034035034036045063000064',
                 'array': [1,2,9,2,9,1,0,-1,-4,8],
                 'apointer': 1,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [2,3,4,5,6,7,8,9],
                 'output': [237,-4,0],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
0690000700050000720730760751400081410131420080040052010002020002030002040002050\
0020600020700020800020900021000018931000030900030800030700030600030500030400030\
3000302000301000211000212000213000214000215000216000217000218000219000220000320\
0003190003180003170003160003150003140003130003120003110002210002220002230002240\
0022500022600022700022800022900023000033000032900032800032700032600032500032400\
0323000322000321000231000232000233000234000235000236000237000238000239000240000\
3400003390003380003370003360003350003340003330003320003310002410002420002430002\
4400024500024600024700024800024900025000035000034900034800034700034600034500034\
4000343000342000341000251000252000253000254000255000256000257000258000259000260\
0003600003590003580003570003560003550003540003530003520003510002610002620002630\
0026400026500026600026700026800026900027000037000036900036800036700036600036500\
0364000363000362000361000271000272000273000274000275000276000277000278000279000\
2800003800003790003780003770003760003750003740003730003720003710002810002820002\
8300028400028500028600028700028800028900029000039000038900038800038700038600038\
5000384000383000382000381000291000292000293000294000295000296000297000298000299\
0000003990003980003970003960003950003940003930003920003910000070200230000000240\
01002003004005000025001002003004005006007008009010000034035034036045063000064',
                 'spointer': 1704},
            # Testing 144, 145
            39: {'in_source': '000000000144145',
                 'array': [1,2,9,2,9,1,0,-1,-4,8],
                 'apointer': 4,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [2,3,4,5,6,7,8,9],
                 'output': [237,-4,0],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
0690000700050000720730760751400081410131420080040052010002020002030002040002050\
0020600020700020800020900021000018931000030900030800030700030600030500030400030\
3000302000301000211000212000213000214000215000216000217000218000219000220000320\
0003190003180003170003160003150003140003130003120003110002210002220002230002240\
0022500022600022700022800022900023000033000032900032800032700032600032500032400\
0323000322000321000231000232000233000234000235000236000237000238000239000240000\
3400003390003380003370003360003350003340003330003320003310002410002420002430002\
4400024500024600024700024800024900025000035000034900034800034700034600034500034\
4000343000342000341000251000252000253000254000255000256000257000258000259000260\
0003600003590003580003570003560003550003540003530003520003510002610002620002630\
0026400026500026600026700026800026900027000037000036900036800036700036600036500\
0364000363000362000361000271000272000273000274000275000276000277000278000279000\
2800003800003790003780003770003760003750003740003730003720003710002810002820002\
8300028400028500028600028700028800028900029000039000038900038800038700038600038\
5000384000383000382000381000291000292000293000294000295000296000297000298000299\
0000003990003980003970003960003950003940003930003920003910000070200230000000240\
0100200300400500002500100200300400500600700800901000003403503403604506300006400\
0000000144145',
                 'spointer': 1719},
            # Testing 146, 147, 156, 157, 165
            40: {'in_source': '146147157156165',
                 'array': [-0.2,-0.4,-1.8,-0.4,-1.6,-0.2,0,0.2,0.8,-1.6],
                 'apointer': 4,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [2,3,4,5,6,7,8,9],
                 'output': [237,-4,0],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
0690000700050000720730760751400081410131420080040052010002020002030002040002050\
0020600020700020800020900021000018931000030900030800030700030600030500030400030\
3000302000301000211000212000213000214000215000216000217000218000219000220000320\
0003190003180003170003160003150003140003130003120003110002210002220002230002240\
0022500022600022700022800022900023000033000032900032800032700032600032500032400\
0323000322000321000231000232000233000234000235000236000237000238000239000240000\
3400003390003380003370003360003350003340003330003320003310002410002420002430002\
4400024500024600024700024800024900025000035000034900034800034700034600034500034\
4000343000342000341000251000252000253000254000255000256000257000258000259000260\
0003600003590003580003570003560003550003540003530003520003510002610002620002630\
0026400026500026600026700026800026900027000037000036900036800036700036600036500\
0364000363000362000361000271000272000273000274000275000276000277000278000279000\
2800003800003790003780003770003760003750003740003730003720003710002810002820002\
8300028400028500028600028700028800028900029000039000038900038800038700038600038\
5000384000383000382000381000291000292000293000294000295000296000297000298000299\
0000003990003980003970003960003950003940003930003920003910000070200230000000240\
0100200300400500002500100200300400500600700800901000003403503403604506300006400\
0000000144145146147157156165',
                 'spointer': 1734},
            # Testing 146, 147, 148, 149, 150, 151, 152, 153, 154, 155, 166, 
            # 167, 168, 169, 170, 171
            41: {'restart': True,
                 'in_source': '020000020000020000020000020166171167170168169146\
147004148149000152154000151000153080150155',
                 'forcedinarray': [1,4,9,16,25,1,2,3,4,5],
                 'array': [0.5,2,4.5,14,4.95,1.75,38.2,1.5,2,2.5], 
                 'apointer': 6,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [1,4,9,16,25],
                 'out_source': '02000002000002000002000002016617116717016816914\
6147004148149000152154000151000153080150155',
                 'spointer': 90},
            # Testing 158, 159, 160, 196, 197, 198
            42: {'in_source': '197198196158159160080',
                 'forcedinarray': [1,4,9,16,25,1,2,3,4,5],
                 'array': [5,20,45,140,49.5,17.5,39,15,20,25], 
                 'apointer': 6,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [1,4,9,16,25],
                 'out_source': '02000002000002000002000002016617116717016816914\
6147004148149000152154000151000153080150155197198196158159160080',
                 'spointer': 108},
            # Testing 074, 078, 079
            43: {'in_source': '000000000000074078000079',
                 'forcedinarray': [1,4,9,16,25,1,2,3,4,5],
                 'array': [1,9,45,140,49.5,17.5,39,15,20,25], 
                 'apointer': 1,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [1,4,9,16,25],
                 'out_source': '02000002000002000002000002016617116717016816914\
6147004148149000152154000151000153080150155197198196158159160080000000000000074\
078000079',
                 'spointer': 135},
            # Testing 037, 039, 040, 041, 110, 111, 114
            44: {'in_source': '037000039000040041110111080114',
                 'forcedinarray': [1,4,9,16,25,1,2,3,4,5],
                 'array': [1,25,1,24,49.5,17.5,39,15,20,25], 
                 'apointer': 3,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [9,16],
                 'out_source': '02000002000002000002000002016617116717016816914\
6147004148149000152154000151000153080150155197198196158159160080000000000000074\
078000079037000039000040041110111080114',
                 'spointer': 165},
            # Testing 123, 124, 125, 126, 127, 128
            45: {'in_source': '123000124125000126000127000128',
                 'forcedinarray': [1,4,9,16,25,1,2,3,4,5],
                 'array': [1,25,1,1,0,1,0,0,20,25], 
                 'apointer': 7,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [9,16],
                 'out_source': '02000002000002000002000002016617116717016816914\
6147004148149000152154000151000153080150155197198196158159160080000000000000074\
078000079037000039000040041110111080114123000124125000126000127000128',
                 'spointer': 195},
            # Testing 143, 161, 162, 163, 164
            46: {'in_source': '161000000163000000164005162004143',
                 'forcedinarray': [1,4,9,16,25,1,2,3,4,5],
                 'array': [1,0,1,0,20,25,0,1,25,1],
                 'apointer': 7,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [9,16],
                 'out_source': '02000002000002000002000002016617116717016816914\
6147004148149000152154000151000153080150155197198196158159160080000000000000074\
0780000790370000390000400411101110801141230001241250001260001270001281610000001\
63000000164005162004143',
                 'spointer': 228},
            # Testing 014, 015 (single loop)
            47: {'restart': True,
                 'in_source': '008008008008008008008008008008014000008008008008\
008004020011015',
                 'forcedinarray': [0,0,0,0,0,0,0,0,0,0],
                 'array': [0,50,0,0,0,0,0,0,0,0],
                 'apointer': 0,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [10,9,8,7,6,5,4,3,2,1],
                 'out_source': '008008008008008008008008008008014000008008008008\
008004020011015',
                 'spointer': 63},
            # Testing 014, 015 (double loop)
            48: {'restart': True,
                 'in_source': '0080080080080080140000080080080080140000080080080\
20004011015004011015001009',
                 'forcedinarray': [0,0,0,0,0,0,0,0,0,0],
                 'array': [0,0,60,0,0,5,0,0,0,0],
                 'apointer': 5,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [3,6,9,12,15,18,21,24,27,30,33,36,39,42,45,48,51,54,
                            57,60],
                 'out_source': '008008008008008014000008008008008014000008008008\
020004011015004011015001009',
                 'spointer': 75},
            # Staging for testing instructions 501 to 599 (Test 50 to 59)
            # Test 49 is the same as Test 37
            49: {'restart': True,
                 'in_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
0690000700050000720730760751400081410131420080040052010002020002030002040002050\
0020600020700020800020900021000018931000030900030800030700030600030500030400030\
3000302000301000211000212000213000214000215000216000217000218000219000220000320\
0003190003180003170003160003150003140003130003120003110002210002220002230002240\
0022500022600022700022800022900023000033000032900032800032700032600032500032400\
0323000322000321000231000232000233000234000235000236000237000238000239000240000\
3400003390003380003370003360003350003340003330003320003310002410002420002430002\
4400024500024600024700024800024900025000035000034900034800034700034600034500034\
4000343000342000341000251000252000253000254000255000256000257000258000259000260\
0003600003590003580003570003560003550003540003530003520003510002610002620002630\
0026400026500026600026700026800026900027000037000036900036800036700036600036500\
0364000363000362000361000271000272000273000274000275000276000277000278000279000\
2800003800003790003780003770003760003750003740003730003720003710002810002820002\
8300028400028500028600028700028800028900029000039000038900038800038700038600038\
5000384000383000382000381000291000292000293000294000295000296000297000298000299\
000000399000398000397000396000395000394000393000392000391',
                 'array': [5,41,9,2,9,1,0,-1,-4,8],
                 'apointer': 8,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
0690000700050000720730760751400081410131420080040052010002020002030002040002050\
0020600020700020800020900021000018931000030900030800030700030600030500030400030\
3000302000301000211000212000213000214000215000216000217000218000219000220000320\
0003190003180003170003160003150003140003130003120003110002210002220002230002240\
0022500022600022700022800022900023000033000032900032800032700032600032500032400\
0323000322000321000231000232000233000234000235000236000237000238000239000240000\
3400003390003380003370003360003350003340003330003320003310002410002420002430002\
4400024500024600024700024800024900025000035000034900034800034700034600034500034\
4000343000342000341000251000252000253000254000255000256000257000258000259000260\
0003600003590003580003570003560003550003540003530003520003510002610002620002630\
0026400026500026600026700026800026900027000037000036900036800036700036600036500\
0364000363000362000361000271000272000273000274000275000276000277000278000279000\
2800003800003790003780003770003760003750003740003730003720003710002810002820002\
8300028400028500028600028700028800028900029000039000038900038800038700038600038\
5000384000383000382000381000291000292000293000294000295000296000297000298000299\
000000399000398000397000396000395000394000393000392000391',
                 'spointer': 1605,
                 'register': None},
            # Testing 501, 502, 503, 504, 505, 506, 507, 508, 509, 510
            50: {'in_source': '501502503504505506507508509510',
                 'array': [5,41,9,2,9,1,0,-1,-4,8],
                 'apointer': 8,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
0690000700050000720730760751400081410131420080040052010002020002030002040002050\
0020600020700020800020900021000018931000030900030800030700030600030500030400030\
3000302000301000211000212000213000214000215000216000217000218000219000220000320\
0003190003180003170003160003150003140003130003120003110002210002220002230002240\
0022500022600022700022800022900023000033000032900032800032700032600032500032400\
0323000322000321000231000232000233000234000235000236000237000238000239000240000\
3400003390003380003370003360003350003340003330003320003310002410002420002430002\
4400024500024600024700024800024900025000035000034900034800034700034600034500034\
4000343000342000341000251000252000253000254000255000256000257000258000259000260\
0003600003590003580003570003560003550003540003530003520003510002610002620002630\
0026400026500026600026700026800026900027000037000036900036800036700036600036500\
0364000363000362000361000271000272000273000274000275000276000277000278000279000\
2800003800003790003780003770003760003750003740003730003720003710002810002820002\
8300028400028500028600028700028800028900029000039000038900038800038700038600038\
5000384000383000382000381000291000292000293000294000295000296000297000298000299\
0000003990003980003970003960003950003940003930003920003915015025035045055065075\
08509510',
                 'spointer': 1635,
                 'register': None},
            # Testing 511, 512, 513, 514, 515, 516, 517, 518, 519, 520
            51: {'in_source': '511512513514515516517518519520',
                 'array': [5,41,9,2,9,1,0,-1,-4,8],
                 'apointer': 8,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
0690000700050000720730760751400081410131420080040052010002020002030002040002050\
0020600020700020800020900021000018931000030900030800030700030600030500030400030\
3000302000301000211000212000213000214000215000216000217000218000219000220000320\
0003190003180003170003160003150003140003130003120003110002210002220002230002240\
0022500022600022700022800022900023000033000032900032800032700032600032500032400\
0323000322000321000231000232000233000234000235000236000237000238000239000240000\
3400003390003380003370003360003350003340003330003320003310002410002420002430002\
4400024500024600024700024800024900025000035000034900034800034700034600034500034\
4000343000342000341000251000252000253000254000255000256000257000258000259000260\
0003600003590003580003570003560003550003540003530003520003510002610002620002630\
0026400026500026600026700026800026900027000037000036900036800036700036600036500\
0364000363000362000361000271000272000273000274000275000276000277000278000279000\
2800003800003790003780003770003760003750003740003730003720003710002810002820002\
8300028400028500028600028700028800028900029000039000038900038800038700038600038\
5000384000383000382000381000291000292000293000294000295000296000297000298000299\
0000003990003980003970003960003950003940003930003920003915015025035045055065075\
08509510511512513514515516517518519520',
                 'spointer': 1665,
                 'register': None},
            # Testing 521, 522, 523, 524, 525, 526, 527, 528, 529, 530
            52: {'in_source': '521522523524525526527528529530',
                 'array': [5,41,9,2,9,1,0,-1,-4,8],
                 'apointer': 8,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
0690000700050000720730760751400081410131420080040052010002020002030002040002050\
0020600020700020800020900021000018931000030900030800030700030600030500030400030\
3000302000301000211000212000213000214000215000216000217000218000219000220000320\
0003190003180003170003160003150003140003130003120003110002210002220002230002240\
0022500022600022700022800022900023000033000032900032800032700032600032500032400\
0323000322000321000231000232000233000234000235000236000237000238000239000240000\
3400003390003380003370003360003350003340003330003320003310002410002420002430002\
4400024500024600024700024800024900025000035000034900034800034700034600034500034\
4000343000342000341000251000252000253000254000255000256000257000258000259000260\
0003600003590003580003570003560003550003540003530003520003510002610002620002630\
0026400026500026600026700026800026900027000037000036900036800036700036600036500\
0364000363000362000361000271000272000273000274000275000276000277000278000279000\
2800003800003790003780003770003760003750003740003730003720003710002810002820002\
8300028400028500028600028700028800028900029000039000038900038800038700038600038\
5000384000383000382000381000291000292000293000294000295000296000297000298000299\
0000003990003980003970003960003950003940003930003920003915015025035045055065075\
08509510511512513514515516517518519520521522523524525526527528529530',
                 'spointer': 1695,
                 'register': None},
            # Testing 531, 532, 533, 534, 535, 536, 537, 538, 539, 540
            53: {'in_source': '531532533534535536537538539540',
                 'array': [5,41,9,2,9,1,0,-1,-4,8],
                 'apointer': 8,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
0690000700050000720730760751400081410131420080040052010002020002030002040002050\
0020600020700020800020900021000018931000030900030800030700030600030500030400030\
3000302000301000211000212000213000214000215000216000217000218000219000220000320\
0003190003180003170003160003150003140003130003120003110002210002220002230002240\
0022500022600022700022800022900023000033000032900032800032700032600032500032400\
0323000322000321000231000232000233000234000235000236000237000238000239000240000\
3400003390003380003370003360003350003340003330003320003310002410002420002430002\
4400024500024600024700024800024900025000035000034900034800034700034600034500034\
4000343000342000341000251000252000253000254000255000256000257000258000259000260\
0003600003590003580003570003560003550003540003530003520003510002610002620002630\
0026400026500026600026700026800026900027000037000036900036800036700036600036500\
0364000363000362000361000271000272000273000274000275000276000277000278000279000\
2800003800003790003780003770003760003750003740003730003720003710002810002820002\
8300028400028500028600028700028800028900029000039000038900038800038700038600038\
5000384000383000382000381000291000292000293000294000295000296000297000298000299\
0000003990003980003970003960003950003940003930003920003915015025035045055065075\
0850951051151251351451551651751851952052152252352452552652752852953053153253353\
4535536537538539540',
                 'spointer': 1725,
                 'register': None},
            # Testing 541, 542, 543, 544, 545, 546, 547, 548, 549, 550
            54: {'in_source': '541542543544545546547548549550',
                 'array': [5,41,9,2,9,1,0,-1,-4,8],
                 'apointer': 8,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
0690000700050000720730760751400081410131420080040052010002020002030002040002050\
0020600020700020800020900021000018931000030900030800030700030600030500030400030\
3000302000301000211000212000213000214000215000216000217000218000219000220000320\
0003190003180003170003160003150003140003130003120003110002210002220002230002240\
0022500022600022700022800022900023000033000032900032800032700032600032500032400\
0323000322000321000231000232000233000234000235000236000237000238000239000240000\
3400003390003380003370003360003350003340003330003320003310002410002420002430002\
4400024500024600024700024800024900025000035000034900034800034700034600034500034\
4000343000342000341000251000252000253000254000255000256000257000258000259000260\
0003600003590003580003570003560003550003540003530003520003510002610002620002630\
0026400026500026600026700026800026900027000037000036900036800036700036600036500\
0364000363000362000361000271000272000273000274000275000276000277000278000279000\
2800003800003790003780003770003760003750003740003730003720003710002810002820002\
8300028400028500028600028700028800028900029000039000038900038800038700038600038\
5000384000383000382000381000291000292000293000294000295000296000297000298000299\
0000003990003980003970003960003950003940003930003920003915015025035045055065075\
0850951051151251351451551651751851952052152252352452552652752852953053153253353\
4535536537538539540541542543544545546547548549550',
                 'spointer': 1755,
                 'register': None},
            # Testing 551, 552, 553, 554, 555, 556, 557, 558, 559, 560
            55: {'in_source': '551552553554555556557558559560',
                 'array': [5,41,9,2,9,1,0,-1,-4,8],
                 'apointer': 8,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
0690000700050000720730760751400081410131420080040052010002020002030002040002050\
0020600020700020800020900021000018931000030900030800030700030600030500030400030\
3000302000301000211000212000213000214000215000216000217000218000219000220000320\
0003190003180003170003160003150003140003130003120003110002210002220002230002240\
0022500022600022700022800022900023000033000032900032800032700032600032500032400\
0323000322000321000231000232000233000234000235000236000237000238000239000240000\
3400003390003380003370003360003350003340003330003320003310002410002420002430002\
4400024500024600024700024800024900025000035000034900034800034700034600034500034\
4000343000342000341000251000252000253000254000255000256000257000258000259000260\
0003600003590003580003570003560003550003540003530003520003510002610002620002630\
0026400026500026600026700026800026900027000037000036900036800036700036600036500\
0364000363000362000361000271000272000273000274000275000276000277000278000279000\
2800003800003790003780003770003760003750003740003730003720003710002810002820002\
8300028400028500028600028700028800028900029000039000038900038800038700038600038\
5000384000383000382000381000291000292000293000294000295000296000297000298000299\
0000003990003980003970003960003950003940003930003920003915015025035045055065075\
0850951051151251351451551651751851952052152252352452552652752852953053153253353\
4535536537538539540541542543544545546547548549550551552553554555556557558559560',
                 'spointer': 1785,
                 'register': None},
            # Testing 561, 562, 563, 564, 565, 566, 567, 568, 569, 570
            56: {'in_source': '561562563564565566567568569570',
                 'array': [5,41,9,2,9,1,0,-1,-4,8],
                 'apointer': 8,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
0690000700050000720730760751400081410131420080040052010002020002030002040002050\
0020600020700020800020900021000018931000030900030800030700030600030500030400030\
3000302000301000211000212000213000214000215000216000217000218000219000220000320\
0003190003180003170003160003150003140003130003120003110002210002220002230002240\
0022500022600022700022800022900023000033000032900032800032700032600032500032400\
0323000322000321000231000232000233000234000235000236000237000238000239000240000\
3400003390003380003370003360003350003340003330003320003310002410002420002430002\
4400024500024600024700024800024900025000035000034900034800034700034600034500034\
4000343000342000341000251000252000253000254000255000256000257000258000259000260\
0003600003590003580003570003560003550003540003530003520003510002610002620002630\
0026400026500026600026700026800026900027000037000036900036800036700036600036500\
0364000363000362000361000271000272000273000274000275000276000277000278000279000\
2800003800003790003780003770003760003750003740003730003720003710002810002820002\
8300028400028500028600028700028800028900029000039000038900038800038700038600038\
5000384000383000382000381000291000292000293000294000295000296000297000298000299\
0000003990003980003970003960003950003940003930003920003915015025035045055065075\
0850951051151251351451551651751851952052152252352452552652752852953053153253353\
4535536537538539540541542543544545546547548549550551552553554555556557558559560\
561562563564565566567568569570',
                 'spointer': 1815,
                 'register': None},
            # Testing 571, 572, 573, 574, 575, 576, 577, 578, 579, 580
            57: {'in_source': '571572573574575576577578579580',
                 'array': [5,41,9,2,9,1,0,-1,-4,8],
                 'apointer': 8,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
0690000700050000720730760751400081410131420080040052010002020002030002040002050\
0020600020700020800020900021000018931000030900030800030700030600030500030400030\
3000302000301000211000212000213000214000215000216000217000218000219000220000320\
0003190003180003170003160003150003140003130003120003110002210002220002230002240\
0022500022600022700022800022900023000033000032900032800032700032600032500032400\
0323000322000321000231000232000233000234000235000236000237000238000239000240000\
3400003390003380003370003360003350003340003330003320003310002410002420002430002\
4400024500024600024700024800024900025000035000034900034800034700034600034500034\
4000343000342000341000251000252000253000254000255000256000257000258000259000260\
0003600003590003580003570003560003550003540003530003520003510002610002620002630\
0026400026500026600026700026800026900027000037000036900036800036700036600036500\
0364000363000362000361000271000272000273000274000275000276000277000278000279000\
2800003800003790003780003770003760003750003740003730003720003710002810002820002\
8300028400028500028600028700028800028900029000039000038900038800038700038600038\
5000384000383000382000381000291000292000293000294000295000296000297000298000299\
0000003990003980003970003960003950003940003930003920003915015025035045055065075\
0850951051151251351451551651751851952052152252352452552652752852953053153253353\
4535536537538539540541542543544545546547548549550551552553554555556557558559560\
561562563564565566567568569570571572573574575576577578579580',
                 'spointer': 1845,
                 'register': None},
            # Testing 581, 582, 583, 584, 585, 586, 587, 588, 589, 590
            58: {'in_source': '581582583584585586587588589590',
                 'array': [5,41,9,2,9,1,0,-1,-4,8],
                 'apointer': 8,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
0690000700050000720730760751400081410131420080040052010002020002030002040002050\
0020600020700020800020900021000018931000030900030800030700030600030500030400030\
3000302000301000211000212000213000214000215000216000217000218000219000220000320\
0003190003180003170003160003150003140003130003120003110002210002220002230002240\
0022500022600022700022800022900023000033000032900032800032700032600032500032400\
0323000322000321000231000232000233000234000235000236000237000238000239000240000\
3400003390003380003370003360003350003340003330003320003310002410002420002430002\
4400024500024600024700024800024900025000035000034900034800034700034600034500034\
4000343000342000341000251000252000253000254000255000256000257000258000259000260\
0003600003590003580003570003560003550003540003530003520003510002610002620002630\
0026400026500026600026700026800026900027000037000036900036800036700036600036500\
0364000363000362000361000271000272000273000274000275000276000277000278000279000\
2800003800003790003780003770003760003750003740003730003720003710002810002820002\
8300028400028500028600028700028800028900029000039000038900038800038700038600038\
5000384000383000382000381000291000292000293000294000295000296000297000298000299\
0000003990003980003970003960003950003940003930003920003915015025035045055065075\
0850951051151251351451551651751851952052152252352452552652752852953053153253353\
4535536537538539540541542543544545546547548549550551552553554555556557558559560\
5615625635645655665675685695705715725735745755765775785795805815825835845855865\
87588589590',
                 'spointer': 1875,
                 'register': None},
            # Testing 591, 592, 593, 594, 595, 596, 597, 598, 599, 600
            59: {'in_source': '591592593594595596597598599600',
                 'array': [5,41,9,2,9,1,0,-1,-4,8],
                 'apointer': 8,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
0690000700050000720730760751400081410131420080040052010002020002030002040002050\
0020600020700020800020900021000018931000030900030800030700030600030500030400030\
3000302000301000211000212000213000214000215000216000217000218000219000220000320\
0003190003180003170003160003150003140003130003120003110002210002220002230002240\
0022500022600022700022800022900023000033000032900032800032700032600032500032400\
0323000322000321000231000232000233000234000235000236000237000238000239000240000\
3400003390003380003370003360003350003340003330003320003310002410002420002430002\
4400024500024600024700024800024900025000035000034900034800034700034600034500034\
4000343000342000341000251000252000253000254000255000256000257000258000259000260\
0003600003590003580003570003560003550003540003530003520003510002610002620002630\
0026400026500026600026700026800026900027000037000036900036800036700036600036500\
0364000363000362000361000271000272000273000274000275000276000277000278000279000\
2800003800003790003780003770003760003750003740003730003720003710002810002820002\
8300028400028500028600028700028800028900029000039000038900038800038700038600038\
5000384000383000382000381000291000292000293000294000295000296000297000298000299\
0000003990003980003970003960003950003940003930003920003915015025035045055065075\
0850951051151251351451551651751851952052152252352452552652752852953053153253353\
4535536537538539540541542543544545546547548549550551552553554555556557558559560\
5615625635645655665675685695705715725735745755765775785795805815825835845855865\
87588589590591592593594595596597598599600',
                 'spointer': 1905,
                 'register': None},
            # Staging for register manipulations
            # Test 60 is the same as Tests 37 and 49
            60: {'restart': True,
                 'in_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
0690000700050000720730760751400081410131420080040052010002020002030002040002050\
0020600020700020800020900021000018931000030900030800030700030600030500030400030\
3000302000301000211000212000213000214000215000216000217000218000219000220000320\
0003190003180003170003160003150003140003130003120003110002210002220002230002240\
0022500022600022700022800022900023000033000032900032800032700032600032500032400\
0323000322000321000231000232000233000234000235000236000237000238000239000240000\
3400003390003380003370003360003350003340003330003320003310002410002420002430002\
4400024500024600024700024800024900025000035000034900034800034700034600034500034\
4000343000342000341000251000252000253000254000255000256000257000258000259000260\
0003600003590003580003570003560003550003540003530003520003510002610002620002630\
0026400026500026600026700026800026900027000037000036900036800036700036600036500\
0364000363000362000361000271000272000273000274000275000276000277000278000279000\
2800003800003790003780003770003760003750003740003730003720003710002810002820002\
8300028400028500028600028700028800028900029000039000038900038800038700038600038\
5000384000383000382000381000291000292000293000294000295000296000297000298000299\
000000399000398000397000396000395000394000393000392000391',
                 'array': [5,41,9,2,9,1,0,-1,-4,8],
                 'apointer': 8,
                 'forcedindata': [1,2,3,4,5,6,7,8,9],
                 'inputdata': [1,2,3,4,5,6,7,8,9],
                 'output': [237],
                 'out_source': '00800800800800000001101101100400400400401101101\
1011011000000000008008008001009002005010043003044006012032013033004004004061004\
0620160180170190460460811330000840000850000860000970000980041200001210001220040\
0400800400807106506800000811502102204700003804200900411608007708700908908809009\
4108096109117105095080000000091093099101092100102103104112113105080066000067000\
0690000700050000720730760751400081410131420080040052010002020002030002040002050\
0020600020700020800020900021000018931000030900030800030700030600030500030400030\
3000302000301000211000212000213000214000215000216000217000218000219000220000320\
0003190003180003170003160003150003140003130003120003110002210002220002230002240\
0022500022600022700022800022900023000033000032900032800032700032600032500032400\
0323000322000321000231000232000233000234000235000236000237000238000239000240000\
3400003390003380003370003360003350003340003330003320003310002410002420002430002\
4400024500024600024700024800024900025000035000034900034800034700034600034500034\
4000343000342000341000251000252000253000254000255000256000257000258000259000260\
0003600003590003580003570003560003550003540003530003520003510002610002620002630\
0026400026500026600026700026800026900027000037000036900036800036700036600036500\
0364000363000362000361000271000272000273000274000275000276000277000278000279000\
2800003800003790003780003770003760003750003740003730003720003710002810002820002\
8300028400028500028600028700028800028900029000039000038900038800038700038600038\
5000384000383000382000381000291000292000293000294000295000296000297000298000299\
000000399000398000397000396000395000394000393000392000391',
                 'spointer': 1605,
                 'register': None},
           }


def comparator(data, result):
    if data == result: return 'PASSED:'
    else: return 'FAILED:'
    
full_source = ''
source = ''

tests = testdata.keys()
# ensure that tests will be executed in ascending case number (1 to N)
tests.sort() 
for t in tests:
    # ----------------------------
    # ------- PREPARE TEST -------
    # ----------------------------
    
    # Step 1: Check to concatenate source or restart source
    try: 
        if testdata[t]['restart'] == True:
            full_source = full_source + source
            source = ''
    except KeyError: pass
     
    isource = source + testdata[t]['in_source']
    
    # Step 2: Get input data list from test (if any)
    try: inputdata = testdata[t]['forcedindata']
    except KeyError: inputdata = []
    
    # Step 3: get pre-execution tape (array) from test (if any)
    try: array = testdata[t]['forcedinarray']
    except KeyError: array = [0]*10
    
    # Step 4: Get expected results after execution
    oarray = testdata[t]['array']            # tape (array) after execution
    oapointer = testdata[t]['apointer']      # tape pointer
    oinputdata = testdata[t]['inputdata']    # input data list
    ooutput = testdata[t]['output']          # output list
    osource = testdata[t]['out_source']      # source
    ospointer = testdata[t]['spointer']      # source pointer
    
    # ----------------------------
    # ------- EXECUTE TEST -------
    # ----------------------------
    
    (array, apointer, inputdata, output, source, spointer) = \
        r.interpret(isource, N.ragaraja, 3, inputdata, array, 10)
    
    # ---------------------------------------------------
    # ------- COMPARE EXPECTED RESULTS AFTER TEST -------
    # ---------------------------------------------------
    
    print(' '.join(['Test number:', str(t), 
                    ', Original source code:', str(isource)]))
    print(' '.join(['    ', str(comparator(oarray, array)), 'array.', 
                    'Expected array:', str(oarray),
                    'Actual array:', str(array)]))
    print(' '.join(['    ', str(comparator(oapointer, apointer)), 
                    'array pointer.', 
                    'Expected array pointer:', str(oapointer),
                    'Actual array pointer:', str(apointer)]))
    print(' '.join(['    ', str(comparator(oinputdata, inputdata)), 
                    'input data list.', 
                    'Expected input data list:', str(oinputdata),
                    'Actual input data list:', str(inputdata)]))
    print(' '.join(['    ', str(comparator(ooutput, output)), 
                    'output list.', 
                    'Expected output list:', str(ooutput),
                    'Actual output list:', str(output)]))
    print(' '.join(['    ', str(comparator(osource, source)), 
                    'source code after execution.', 
                    'Expected source:', str(osource),
                    'Actual source:', str(source)]))
    print(' '.join(['    ', str(comparator(ospointer, spointer)), 
                    'source code pointer after execution.', 
                    'Expected source pointer:', str(ospointer),
                    'Actual source pointer:', str(spointer)]))
    if testdata[t].has_key('register'):
        print('Internal registers: ' + str(N.register))
    print('==========================================================')
# -----------------------------
# --------- REGISTERS --------- 
# -----------------------------

print()
print('===== Internal register state =====')
print('Internal registers: ' + str(N.register))

# ---------------------------------------------
# --------- TESTING RANDOM OPERATIONS --------- 
# ---------------------------------------------

print()
print('===== Testing random operations =====')
print('Testing 050 051 052 053 054 055 056 057 058 059 060')
print()
random_source = '050051052053054055056057058059060'
random_source = random_source + random_source + random_source + \
                random_source + random_source + random_source
for x in range(20):
    tape = r.interpret(random_source, N.ragaraja, 3, [], [0]*10, 10)[0]
    print('Test #' + str(x+1) + ', Tape: ' + str(tape))
print('===== End of random operations testing =====')

# ---------------------------
# --------- SUMMARY --------- 
# ---------------------------

isource = full_source + isource + random_source
instruction_set = {}
for i in range(0, len(isource), 3): instruction_set[isource[i:i+3]] = ''
instruction_set = instruction_set.keys()
instruction_set.sort()
print() 
print('Instruction set tested: ')
print(' '.join(instruction_set))
print()
print('Number of instructions tested :' + str(len(instruction_set)))
print()
print('------------------------- END OF TESTS -------------------------')